/*
 *************************************************************************
 *                               Ebert.c                                 *
 *************************************************************************
 * This file contains C code for procedures in my chapters of "Texturing *
 * and Modeling: A Procedural Approach". This software may be used for   *
 * non-profit use as long as the original author is credited.  If you    *
 * find any bugs in this software, send email.                           *
 *************************************************************************
 * Author: David S. Ebert (ebert@cs.umbc.edu)                            *
 * Copyright 1994, 1998 David S. Ebert                                   *
 *************************************************************************
*/

/*
 *************************************************************************
 *                           WRITE_NOISE.C                               *
 *  This procedure generates a noise function file for solid texturing   *
 *                      by David S. Ebert                                *
 *************************************************************************
 * This noise file needs to be read in to your program for the rest of   *
 * the procedures in this file to work.                                  *
 *************************************************************************
 */
#define _USE_MATH_DEFINES
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "ebert.h"

extern "C"
{
    #include "PROCTEXT.H"
}

//#include "NOISE.H"
//#include "macro.h"

#define POW_TABLE_SIZE 10000
#define RAMP_SIZE      200
#define OFFSET_SIZE    1024
#define SWIRL_FRAMES   126 
#define SWIRL          0.04986655               /* 2*PI/126  */
#define SWIRL_AMOUNT   0.04986655               /* 2*PI/126  */
#define DOWN           0.011904762
#define SPEED          .012
#define DOWN_AMOUNT    .0095
#define DOWN_SMOKE     .015
#define RAD1           .10
#define RAD2           .14
#define SWIRL_SMOKE    .062831853              /* 2*PI/100  */
#define SWIRL_FRAMES_SMOKE 100                 

#define PIXEL_SIZE 0.016666667 // 1/60 -- 0.0025 // 1/400

float drand48() {return (float)rand() / RAND_MAX;}
float   noise[RAND_GRID+1][RAND_GRID+1][RAND_GRID+1];
int     frame_num;


void MetaClump::bounds (xyz *minPt, xyz *maxPt, const float &stretch) const {
    minPt->set ((float)INT_MAX);
    maxPt->set ((float)INT_MIN);

    for (MetaPrimVec::const_iterator primIt = prims.begin();
                               primIt != prims.end ();
                               ++primIt) {
        xyz primMin, primMax;
        (*primIt)->bounds (&primMin, &primMax, stretch);

        maxPt->x = max(maxPt->x, primMax.x);
        maxPt->y = max(maxPt->y, primMax.y);
        maxPt->z = max(maxPt->z, primMax.z);

        minPt->x = min(minPt->x, primMin.x);
        minPt->y = min(minPt->y, primMin.y);
        minPt->z = min(minPt->z, primMin.z);
    }
}


void Sphere::bounds (xyz *minPt, xyz *maxPt, const float &stretch) const {
    maxPt->x = pos.x + r * stretch;
    maxPt->y = pos.y + r * stretch;
    maxPt->z = pos.z + r * stretch;

    minPt->x = pos.x - r * stretch;
    minPt->y = pos.y - r * stretch;
    minPt->z = pos.z - r * stretch;
}


float Sphere::density(const xyz &_pt) const {
    // pg 271 of Texturing & Modeling
    // return -(4/9)*(dist^6 / r^6) + (17/9)*(dist^4 / r^4) - (22/9)*(dist^2 / r^2) + 1

    const float fract1 = (float)0.44444444444444444444444444444444; // 4.f / 9.f;
    const float fract2 = (float)1.8888888888888888888888888888889; // 17.f / 9.f;
    const float fract3 = (float)2.4444444444444444444444444444444; // 22.f / 9.f;

    // dist^2, dist^4, dist^6
    float dist2 = (_pt - pos).length2();

    if (dist2 >= r2)
        return 0.f;

    float dist4 = dist2 * dist2;
    float dist6 = dist4 * dist2;

    float part1 = fract1 * (dist6 * r6Inv); // (4/9)*(dist^6 / r^6)
    float part2 = fract2 * (dist4 * r4Inv); // (17/9)*(dist^4 / r^4)
    float part3 = fract3 * (dist2 * r2Inv); // (22/9)*(dist^2 / r^2)

    float Fcub = -part1 + part2 - part3 + 1;

    return 0 >= Fcub ? 0 : weight * Fcub;
}

float Ellipse::density(const xyz &_pt) const {

    // put put in local space
    xyz spherePt (_pt - pos);

    if (spherePt.x > r.x || spherePt.y > r.y || spherePt.z > r.z )
        return 0.f;

    // stretch point as if it were inside a sphere
    spherePt.x *= spherize.x;
    spherePt.y *= spherize.y;
    spherePt.z *= spherize.z;


    // pg 271 of Texturing & Modeling
    // return -(4/9)*(dist^6 / r^6) + (17/9)*(dist^4 / r^4) - (22/9)*(dist^2 / r^2) + 1

    const float fract1 = (float)0.44444444444444444444444444444444; // 4.f / 9.f;
    const float fract2 = (float)1.8888888888888888888888888888889; // 17.f / 9.f;
    const float fract3 = (float)2.4444444444444444444444444444444; // 22.f / 9.f;

    // dist^2, dist^4, dist^6
    float dist2 = spherePt.length2();

    if (dist2 >= rmax2)
        return 0.f;

    float dist4 = dist2 * dist2;
    float dist6 = dist4 * dist2;

    float part1 = fract1 * (dist6 * rmax6Inv); // (4/9)*(dist^6 / r^6)
    float part2 = fract2 * (dist4 * rmax4Inv); // (17/9)*(dist^4 / r^4)
    float part3 = fract3 * (dist2 * rmax2Inv); // (22/9)*(dist^2 / r^2)

    float Fcub = -part1 + part2 - part3 + 1;

    return 0 >= Fcub ? 0 : weight * Fcub;
}

void Ellipse::bounds (xyz *minPt, xyz *maxPt, const float &stretch) const {
    maxPt->x = pos.x + r.x * stretch;
    maxPt->y = pos.y + r.y * stretch;
    maxPt->z = pos.z + r.z * stretch;

    minPt->x = pos.x - r.x * stretch;
    minPt->y = pos.y - r.y * stretch;
    minPt->z = pos.z - r.z * stretch;
}


void generate_noise()
{
  long    i,j, k, ii,jj,kk;
 
  for (i=0; i<RAND_GRID; i++)
    for (j=0; j<RAND_GRID; j++)
      for (k=0; k<RAND_GRID; k++)
	{
	  noise[i][j][k] = (float)drand48();
	}
  
  // This is a hack, but it works. Remember this is only done once.

  for (i=0; i<RAND_GRID+1; i++)
    for (j=0; j<RAND_GRID+1; j++)
      for (k=0; k<RAND_GRID+1; k++)
	{
	  ii = (i == RAND_GRID)? 0:  i;
	  jj = (j == RAND_GRID)? 0:  j;
	  kk = (k == RAND_GRID)? 0:  k;
	  noise[i][j][k] = noise[ii][jj][kk];
	}
}

/*
 ***********************************************************************
 * The procedures below use the noise file written by the above        *
 * write_noise program. Read the noise file into a variable called     *
 * noise. These procedures also assume that the renderer has a global  *
 * int variable frame_num  that contains the current frame number.     *
 ***********************************************************************
 */


/*
 ***********************************************************************
 *                            Calc_noise                               *
 ***********************************************************************
 * This is basically how the trilinear interpolation works. I lerp     *
 * down the left front edge of the cube, then the right front edge of  *
 * the cube(p_l, p_r). Then I lerp down the left back and right back   *
 * edges of the  cube (p_l2, p_r2). Then I lerp across the front face  *
 * between p_l  and p_r (p_face1). The I lerp across the back face     *
 * between p_l2 and p_r2 (p_face2). Now I lerp along the line between  *
 * p_face1 and  p_face2.                                               *
 ***********************************************************************
 */

float calc_noise(xyz pnt)
{
  float  t1;
  float  p_l,p_l2,    /* value lerped down left side of face1 & face 2 */
         p_r,p_r2,    /* value lerped down left side of face1 & face 2 */
         p_face1,     /* value lerped across face 1 (x-y plane ceil of z) */
         p_face2,     /* value lerped across face 2 (x-y plane floor of z) */
         p_final;     /* value lerped through cube (in z)                  */

  register int x, y, z, px, py, pz;
  
  px = (int)pnt.x;
  py = (int)pnt.y;
  pz = (int)pnt.z;
  x = px &(RAND_GRID-1); /* make sure the values are in the table           */
  y = py &(RAND_GRID-1); /* Effectively, replicates the table thought space */
  z = pz &(RAND_GRID-1);
  
  t1 = pnt.y - py;
  p_l  = noise[x  ][y][z+1] + t1 * (noise[x  ][y+1][z+1] - noise[x  ][y][z+1]);  
  p_r  = noise[x+1][y][z+1] + t1 * (noise[x+1][y+1][z+1] - noise[x+1][y][z+1]);
  p_l2 = noise[x  ][y][z  ] + t1 * (noise[x  ][y+1][z  ] - noise[x  ][y][z  ]);     
  p_r2 = noise[x+1][y][z  ] + t1 * (noise[x+1][y+1][z  ] - noise[x+1][y][z  ]);
  
  t1 = pnt.x - px; 
  p_face1 = p_l  + t1 * (p_r  - p_l );
  p_face2 = p_l2 + t1 * (p_r2 - p_l2);
  
  t1 = pnt.z - pz;
  p_final = p_face2 + t1 * (p_face1 - p_face2);
  
  return(p_final);
}


/*
 **********************************************************************
 *                   TURBULENCE                                       *
 **********************************************************************
 */

// optimize
float turbulence(xyz pnt)//, float pixel_size)
{
    float t, scale;

    t = 0;
    for(scale = 1.0; scale > PIXEL_SIZE; scale *= 0.5) {
        pnt /= scale;
        t += calc_noise(pnt) * scale ; 
    }
    return t;
}

/*
 **********************************************************************
 * Cumulus Cloud function
 **********************************************************************
 */

// return density
float cumulus_density (xyz pnt,      // location of point in cloud space 
                       xyz pnt_w,    // location of point in world space 
                       const MetaClump &vol,// volume
                       float param1,  // density scaling 
                       float param2,  // exponent in pow() function
                       float param3,  // blending function value between meta-ball density and product of turbulence density. larger means less turbulent
                       float param4)  // turbulence in displacing primitive before evaluation of implicit primitives
{
    /*
    float new_turbulence();     // my turbulence function 
    float peachey_noise();      // Darwyn Peachey's noise function 
    float metaball_evaluate();  // function for evaluating metaball primitives
    */
    float mdens,                // metaball density value 
          turb,                 // turbulence in lookup
          turb_amount,          // turb * param4
          peach;                // Peachey noise value 
    xyz path;                // path for swirling the point
    int frame_num = 30; // arbitrary
    static int ncalcd=1;
    static float sin_theta_cloud, cos_theta_cloud, theta,
         path_x, path_y, path_z, scalar_x, scalar_y, scalar_z;

    // calculate values that only depend on the frame number 
    // once per frame
    if(ncalcd)
    {
        ncalcd=0;
        // create gentle swirling in the cloud 
        theta = (frame_num % 600) * .01047196f; // swirling effect 
        cos_theta_cloud = cosf (theta);	
        sin_theta_cloud = sinf (theta);
        path_x = sin_theta_cloud * .005f * frame_num;
        path_y = .01215f * (float)frame_num;
        path_z= sin_theta_cloud * .0035f * frame_num;
        scalar_x = (.5f + (float)frame_num * 0.010f);
        scalar_z = (float)frame_num * .0073f;
    }

    // Add some noise to the point's location 
    peach = vnoise (pnt.x, pnt.y, pnt.z);//peachey_noise (pnt); // Use Darwyn Peachey's noise function 
    pnt.x -= path_x - (peach * scalar_x);
    pnt.y = pnt.y - path_y + (.5f * peach);
    pnt.z += path_z - (peach * scalar_z);

    // Perturb the location of the point before evaluating the implicit
    // primitives.
    turb = turbulence (pnt);//fast_turbulence(pnt);
    turb_amount = param4 * turb;
    pnt_w.x += turb_amount;
    pnt_w.y -= turb_amount;
    pnt_w.z += turb_amount;

    mdens = vol.evaluate (pnt_w);

    float density = param1 * ( (param3 * mdens) + (param3 * turb * mdens) );
    //float density = param1 * ( (param3 * mdens) + (1.0f - (param3 * turb * mdens)) );
    return powf (density, param2);
}


/*
 **********************************************************************
 * Cirrus Cloud Function
 **********************************************************************
 */
/*
void Cirrus(xyz_td pnt,          // location of point in cloud space
	    float *density,
	    float *parms, 
	    xyz_td pnt_w,        // location of point in world space
	    vol_td vol, 
	    xyz_td jet_stream)
{
  float new_turbulence();     // my turbulence function 
  float peachey_noise();      // Darwyn Peachey's noise function 
  float metaball_evaluate();  // function for evaluating the metaball primitives
  float mdens,                // metaball density value 
        turb, turb_amount,    // turbulence amount 
        peach;                // Peachey noise value 
  xyz_td path;                // path for swirling the point 
  extern int frame_num;
  static int ncalcd=1;
  static float sin_theta_cloud, cos_theta_cloud, theta,
         path_x, path_y, path_z, scalar_x, scalar_y, scalar_z;
  
  // calculate values that only depend on the frame number 
  // once per frame
  if(ncalcd)
    {
      ncalcd = 0;
      // create gentle swirling in the cloud 
      theta = (frame_num % 600) * .01047196f; // swirling effect 
      cos_theta_cloud = cosf(theta);	
      sin_theta_cloud = sinf(theta);
      path_x = sin_theta_cloud * .005f * frame_num;
      path_y = .01215f * (float)frame_num;
      path_z= sin_theta_cloud * .0035f * frame_num;
      scalar_x = .5f + (frame_num * 0.010f);
      scalar_z = frame_num * .0073f;
    }

   // Add some noise to the point's location 
   peach = peachey_noise(pnt); // Use Darwyn Peachey's noise function 
   pnt.x -= path_x - (peach * scalar_x);
   pnt.y = pnt.y - path_y + (.5f * peach);
   pnt.z += path_z - peach*scalar_z;

   // Perturb the location of the point before evaluating the implicit
   // primitives.
   turb=turbulence(pnt);//fast_turbulence(pnt);
   turb_amount=parms[4]*turb;
   pnt_w.x += turb_amount;
   pnt_w.y -= turb_amount;
   pnt_w.z += turb_amount;

   // make the jet stream turbulent 
   jet_stream.x += .2f * turb;
   jet_stream.y += .3f * turb;
   jet_stream.z += .25f * turb;
 
   // warp point along the jet stream vector 
   pnt_w = warp(jet_stream, pnt_w);

   mdens = metaball_evaluate(pnt_w.x, pnt_w.y, pnt_w.z, (vol.metaball));

  *density= parms[1]*(parms[3]*mdens + (1.0f - parms[3])*turb*mdens);
  *density= pow(*density, parms[2]);
}
*/
