
/*
 * pbrt source code Copyright(c) 1998-2005 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For educational use only; commercial use expressly forbidden.
 * NO WARRANTY, express or implied, for this software.
 * (See file License.txt for complete license)
 */

#ifndef PBRT_COLOR_H
#define PBRT_COLOR_H
// color.h*
#include "pbrt.h"
// Spectrum Declarations
class COREDLL Spectrum {
public:
	// Spectrum Public Methods
	Spectrum(float v = 0.f) {
		for (int i = 0; i < COLOR_SAMPLES; ++i)
			c[i] = v;
	}
	Spectrum(float cs[COLOR_SAMPLES]) {
		for (int i = 0; i < COLOR_SAMPLES; ++i)
			c[i] = cs[i];
	}
	friend ostream &operator<<(ostream &, const Spectrum &);
	Spectrum &operator+=(const Spectrum &s2) {
		for (int i = 0; i < COLOR_SAMPLES; ++i)
			c[i] += s2.c[i];
		return *this;
	}
	Spectrum operator+(const Spectrum &s2) const {
		Spectrum ret = *this;
		for (int i = 0; i < COLOR_SAMPLES; ++i)
			ret.c[i] += s2.c[i];
		return ret;
	}
	Spectrum operator-(const Spectrum &s2) const {
		Spectrum ret = *this;
		for (int i = 0; i < COLOR_SAMPLES; ++i)
			ret.c[i] -= s2.c[i];
		return ret;
	}
	Spectrum operator/(const Spectrum &s2) const {
		Spectrum ret = *this;
		for (int i = 0; i < COLOR_SAMPLES; ++i)
			ret.c[i] /= s2.c[i];
		return ret;
	}
	Spectrum operator*(const Spectrum &sp) const {
		Spectrum ret = *this;
		for (int i = 0; i < COLOR_SAMPLES; ++i)
			ret.c[i] *= sp.c[i];
		return ret;
	}
	Spectrum &operator*=(const Spectrum &sp) {
		for (int i = 0; i < COLOR_SAMPLES; ++i)
			c[i] *= sp.c[i];
		return *this;
	}
	Spectrum operator*(float a) const {
		Spectrum ret = *this;
		for (int i = 0; i < COLOR_SAMPLES; ++i)
			ret.c[i] *= a;
		return ret;
	}
	Spectrum &operator*=(float a) {
		for (int i = 0; i < COLOR_SAMPLES; ++i)
			c[i] *= a;
		return *this;
	}
	friend inline
	Spectrum operator*(float a, const Spectrum &s) {
		return s * a;
	}
	Spectrum operator/(float a) const {
		return *this * (1.f / a);
	}
	Spectrum &operator/=(float a) {
		float inv = 1.f / a;
		for (int i = 0; i < COLOR_SAMPLES; ++i)
			c[i] *= inv;
		return *this;
	}
	void AddWeighted(float w, const Spectrum &s) {
		for (int i = 0; i < COLOR_SAMPLES; ++i)
			c[i] += w * s.c[i];
	}
	bool operator==(const Spectrum &sp) const {
		for (int i = 0; i < COLOR_SAMPLES; ++i)
			if (c[i] != sp.c[i]) return false;
		return true;
	}
	bool operator!=(const Spectrum &sp) const {
		return !(*this == sp);
	}
	bool Black() const {
		for (int i = 0; i < COLOR_SAMPLES; ++i)
			if (c[i] != 0.) return false;
		return true;
	}
	Spectrum Sqrt() const {
		Spectrum ret;
		for (int i = 0; i < COLOR_SAMPLES; ++i)
			ret.c[i] = sqrtf(c[i]);
		return ret;
	}
	Spectrum Pow(const Spectrum &e) const {
		Spectrum ret;
		for (int i = 0; i < COLOR_SAMPLES; ++i)
			ret.c[i] = c[i] > 0 ? powf(c[i], e.c[i]) : 0.f;
		return ret;
	}
	Spectrum operator-() const {
		Spectrum ret;
		for (int i = 0; i < COLOR_SAMPLES; ++i)
			ret.c[i] = -c[i];
		return ret;
	}
	friend Spectrum Exp(const Spectrum &s) {
		Spectrum ret;
		for (int i = 0; i < COLOR_SAMPLES; ++i)
			ret.c[i] = expf(s.c[i]);
		return ret;
	}
	Spectrum Clamp(float low = 0.f,
	               float high = INFINITY) const {
		Spectrum ret;
		for (int i = 0; i < COLOR_SAMPLES; ++i)
			ret.c[i] = ::Clamp(c[i], low, high);
		return ret;
	}
	bool IsNaN() const {
		for (int i = 0; i < COLOR_SAMPLES; ++i)
			if (isnan(c[i])) return true;
		return false;
	}
	void Print(FILE *f) const {
		for (int i = 0; i < COLOR_SAMPLES; ++i)
			fprintf(f, "%f ", c[i]);
	}
	void XYZ(float xyz[3]) const {
		xyz[0] = xyz[1] = xyz[2] = 0.;
		for (int i = 0; i < COLOR_SAMPLES; ++i) {
			xyz[0] += XWeight[i] * c[i];
			xyz[1] += YWeight[i] * c[i];
			xyz[2] += ZWeight[i] * c[i];
		}
	}
	float y() const {
		float v = 0.;
		for (int i = 0; i < COLOR_SAMPLES; ++i)
			v += YWeight[i] * c[i];
		return v;
	}
	bool operator<(const Spectrum &s2) const {
		return y() < s2.y();
	}
	friend class ParamSet;
	
	// Spectrum Public Data
	static const int CIEstart = 360;
	static const int CIEend = 830;
	static const int nCIE = CIEend-CIEstart+1;
	static const float CIE_X[nCIE];
	static const float CIE_Y[nCIE];
	static const float CIE_Z[nCIE];
private:
	// Spectrum Private Data
	float c[COLOR_SAMPLES];
	static float XWeight[COLOR_SAMPLES];
	static float YWeight[COLOR_SAMPLES];
	static float ZWeight[COLOR_SAMPLES];
	friend Spectrum FromXYZ(float x, float y, float z);
};
// *NEW* Higher sampling rate spectrum for sky calculations
class COREDLL SpectrumHR {
public:
	// Spectrum Public Methods
	SpectrumHR(float v = 0.f) {
		for (int i = 0; i < COLOR_SAMPLES_HR; ++i)
			c[i] = v;
	}
	SpectrumHR(float cs[COLOR_SAMPLES]) {
		for (int i = 0; i < COLOR_SAMPLES_HR; ++i)
			c[i] = cs[i];
	}
	friend ostream &operator<<(ostream &, const SpectrumHR &);
	SpectrumHR &operator+=(const SpectrumHR &s2) {
		for (int i = 0; i < COLOR_SAMPLES_HR; ++i)
			c[i] += s2.c[i];
		return *this;
	}
	SpectrumHR operator+(const SpectrumHR &s2) const {
		SpectrumHR ret = *this;
		for (int i = 0; i < COLOR_SAMPLES_HR; ++i)
			ret.c[i] += s2.c[i];
		return ret;
	}
	SpectrumHR operator-(const SpectrumHR &s2) const {
		SpectrumHR ret = *this;
		for (int i = 0; i < COLOR_SAMPLES_HR; ++i)
			ret.c[i] -= s2.c[i];
		return ret;
	}
	SpectrumHR operator/(const SpectrumHR &s2) const {
		SpectrumHR ret = *this;
		for (int i = 0; i < COLOR_SAMPLES_HR; ++i)
			ret.c[i] /= s2.c[i];
		return ret;
	}
	SpectrumHR operator*(const SpectrumHR &sp) const {
		SpectrumHR ret = *this;
		for (int i = 0; i < COLOR_SAMPLES_HR; ++i)
			ret.c[i] *= sp.c[i];
		return ret;
	}
	SpectrumHR &operator*=(const SpectrumHR &sp) {
		for (int i = 0; i < COLOR_SAMPLES_HR; ++i)
			c[i] *= sp.c[i];
		return *this;
	}
	SpectrumHR operator*(float a) const {
		SpectrumHR ret = *this;
		for (int i = 0; i < COLOR_SAMPLES_HR; ++i)
			ret.c[i] *= a;
		return ret;
	}
	SpectrumHR &operator*=(float a) {
		for (int i = 0; i < COLOR_SAMPLES_HR; ++i)
			c[i] *= a;
		return *this;
	}
	friend inline
	SpectrumHR operator*(float a, const SpectrumHR &s) {
		return s * a;
	}
	SpectrumHR operator/(float a) const {
		return *this * (1.f / a);
	}
	SpectrumHR &operator/=(float a) {
		float inv = 1.f / a;
		for (int i = 0; i < COLOR_SAMPLES_HR; ++i)
			c[i] *= inv;
		return *this;
	}
	void AddWeighted(float w, const SpectrumHR &s) {
		for (int i = 0; i < COLOR_SAMPLES_HR; ++i)
			c[i] += w * s.c[i];
	}
	bool operator==(const SpectrumHR &sp) const {
		for (int i = 0; i < COLOR_SAMPLES_HR; ++i)
			if (c[i] != sp.c[i]) return false;
		return true;
	}
	bool operator!=(const SpectrumHR &sp) const {
		return !(*this == sp);
	}
	bool Black() const {
		for (int i = 0; i < COLOR_SAMPLES_HR; ++i)
			if (c[i] != 0.) return false;
		return true;
	}
	SpectrumHR Sqrt() const {
		SpectrumHR ret;
		for (int i = 0; i < COLOR_SAMPLES_HR; ++i)
			ret.c[i] = sqrtf(c[i]);
		return ret;
	}
	SpectrumHR Pow(const SpectrumHR &e) const {
		SpectrumHR ret;
		for (int i = 0; i < COLOR_SAMPLES_HR; ++i)
			ret.c[i] = c[i] > 0 ? powf(c[i], e.c[i]) : 0.f;
		return ret;
	}
	SpectrumHR operator-() const {
		SpectrumHR ret;
		for (int i = 0; i < COLOR_SAMPLES_HR; ++i)
			ret.c[i] = -c[i];
		return ret;
	}
	friend SpectrumHR Exp(const SpectrumHR &s) {
		SpectrumHR ret;
		for (int i = 0; i < COLOR_SAMPLES_HR; ++i)
			ret.c[i] = expf(s.c[i]);
		return ret;
	}
	SpectrumHR Clamp(float low = 0.f,
	               float high = INFINITY) const {
		SpectrumHR ret;
		for (int i = 0; i < COLOR_SAMPLES_HR; ++i)
			ret.c[i] = ::Clamp(c[i], low, high);
		return ret;
	}
	bool IsNaN() const {
		for (int i = 0; i < COLOR_SAMPLES_HR; ++i)
			if (isnan(c[i])) return true;
		return false;
	}
	void Print(FILE *f) const {
		for (int i = 0; i < COLOR_SAMPLES_HR; ++i)
			fprintf(f, "%f ", c[i]);
	}
	void XYZ(float xyz[3]) const {
		xyz[0] = xyz[1] = xyz[2] = 0.;
		int s = 0;
		int end;
		for (int i = 0; i < COLOR_SAMPLES_HR; ++i) {
			if (i == COLOR_SAMPLES_HR - 1)
				end = Spectrum::CIEend - Spectrum::CIEstart;
			else
				end = (int)sampleWavelengths[i + 1] - wavelengthRange - Spectrum::CIEstart;
			for (;s <= end; ++s) {
				xyz[0] += c[i]*Spectrum::CIE_X[s];
				xyz[1] += c[i]*Spectrum::CIE_Y[s];
				xyz[2] += c[i]*Spectrum::CIE_Z[s];
			}
		}
	}
	float y() const {
		float v = 0.f;
		int s = 0;
		for (int i = 0; i < COLOR_SAMPLES_HR; ++i) {
			for (int j = 0; j < sampleWidth && s < Spectrum::nCIE; ++j, ++s) {
				v += c[i]*Spectrum::CIE_Y[s];
			}
		}
		return v;
	}
	bool operator<(const SpectrumHR &s2) const {
		return y() < s2.y();
	}
	static const float* SampleWavelengths() {
		return sampleWavelengths;
	}
	friend class ParamSet;
static int sampleWidth;
private:
	// SpectrumHR Private Data
	const static float sampleWavelengths[9];// {400.f, 450.f, 500.f, 550.f, 600.f, 650.f, 700.f, 750.f, 800.f};
	const static int wavelengthRange = 25;
	float c[COLOR_SAMPLES_HR];
	
};
#endif // PBRT_COLOR_H
