
/*
 * pbrt source code Copyright(c) 1998-2005 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For educational use only; commercial use expressly forbidden.
 * NO WARRANTY, express or implied, for this software.
 * (See file License.txt for complete license)
 */

// multiple.cpp*
#include "volume.h"
#include "transport.h"
#include "scene.h"

using namespace std;

Vector SampleCosWeightedSphere(const Vector &w, float u1, float u2) {
    float theta = 2*asinf(u1);
    float costheta = cosf(theta);
    float sintheta = sqrtf(1.f - costheta*costheta);
	float phi = 2.f * M_PI * u2;
	Vector v1, v2;
	CoordinateSystem(w, &v1, &v2);
	return SphericalDirection(sintheta, costheta,
		phi, v1, v2, w);
}

float CosWeightedSpherePdf (const Vector &w, const Vector &wp) {
    float theta = acosf (Dot (w, wp));
    return .5f * cosf (theta * .5f) * INV_TWOPI;
}

// multipleScattering Declarations
class MultipleScattering : public VolumeIntegrator {
public:
	// MultipleScattering Public Methods
	MultipleScattering(float ss, int _nScatters, const int *_samplesPerScatter) {
        stepSize = ss;
        nScatters = _nScatters;
        samplesAtScatterI = new int[nScatters];
        memcpy(samplesAtScatterI, _samplesPerScatter, nScatters * sizeof(int));
    }
    ~MultipleScattering() {delete [] samplesAtScatterI;}
	Spectrum Transmittance(const Scene *, const Ray &ray,
		const Sample *sample, float *alpha) const;
	void RequestSamples(Sample *sample, const Scene *scene);
	Spectrum Li(const Scene *, const RayDifferential &ray, const Sample *sample,
                float *alpha) const;
private:
	Spectrum Li_multi(const Scene *, const RayDifferential &ray,
                             const Sample *sample, float *alpha,
                             const int scatterI, const float influence) const;

	// MultipleScattering Private Data
	float stepSize;
	int tauSampleOffset, scatterSampleOffset;
    int nScatters, *samplesAtScatterI;
};
// MultipleScattering Method Definitions
void MultipleScattering::RequestSamples(Sample *sample,
		const Scene *scene) {
	tauSampleOffset = sample->Add1D(1);
	scatterSampleOffset = sample->Add1D(1);
}
Spectrum MultipleScattering::Transmittance(const Scene *scene,
		const Ray &ray, const Sample *sample, float *alpha) const {
	if (!scene->volumeRegion) return Spectrum(1.f);
	float step = sample ? stepSize : 4.f * stepSize;
	float offset = sample ? sample->oneD[tauSampleOffset][0] :
		RandomFloat();
	Spectrum tau = scene->volumeRegion->Tau(ray, step, offset);
	return Exp(-tau);
}
Spectrum MultipleScattering::Li(const Scene *scene,
		const RayDifferential &ray, const Sample *sample,
		float *alpha) const {
    return Li_multi(scene, ray, sample, alpha, 0, 1.f);
}

Spectrum MultipleScattering::Li_multi(const Scene *scene, const RayDifferential &ray,
        const Sample *sample, float *alpha, const int scatterI, const float influence) const {

    VolumeRegion *vr = scene->volumeRegion;
	float t0, t1;
	if (!vr || !vr->IntersectP(ray, &t0, &t1))
        return 0.f;

	// Do single scattering volume integration in _vr_
	Spectrum Lv(0.);

	// Prepare for volume integration stepping
    // step size increases by 2^scatterI
    float stepSizeI = stepSize * (((unsigned int)1) << scatterI); 
	int nSteps = Ceil2Int((t1-t0) / stepSizeI);
	float step = (t1 - t0) / nSteps;
	Spectrum Tr(1.f); // transmission
	Point p = ray(t0), pPrev;
	Vector w = -ray.d;
	if (sample) // initial ray through volume
		t0 += sample->oneD[scatterSampleOffset][0] * step;
	else // shadow ray or multi-scatter
		t0 += RandomFloat() * step;

	// Compute sample patterns for single scattering samples
    // samp[stepI][sampleI][0]: light to sample (if sampling light)
    // samp[stepI][sampleI][1 & 2]: uv on light we're sampling OR uv for phase function
    const int randPerSample = 3;
    int nSamp = nSteps * samplesAtScatterI[scatterI];
	float *samp = (float *)alloca(nSamp * randPerSample * sizeof(float));
	LatinHypercube(samp, nSamp, randPerSample);

    int nLights = scene->lights.size();
	int sampOffsetStep = 0;
	for (int stepI = 0; stepI < nSteps; ++stepI, t0 += step) {
		// Advance to sample at _t0_ and update _T_
		pPrev = p;
		p = ray(t0);
		Spectrum stepTau = vr->Tau(Ray(pPrev, p - pPrev, 0, 1),
			.5f * stepSizeI, RandomFloat());
		Tr *= Exp(-stepTau);

		// Possibly terminate raymarching if transmittance is small
		if (Tr.y() < 1e-3) {
			const float continueProb = .5f;
			if (RandomFloat() > continueProb) break;
			Tr /= continueProb;
		}
        // Compute emission at _p_
		Lv += Tr * vr->Lve(p, w);

		// Compute scattering source term at _p_ (if possible for any light to scatter back)
		Spectrum ss = vr->sigma_s(p, w);
	    if (!ss.Black() && scene->lights.size() > 0) {
            Vector wo;
	        float u1, u2;

            // this is not the final scatter, scatter recursively X times
            if (scatterI < nScatters - 1)
            {
                int sampOffsetSample = 0;
                Spectrum Ld (0.f);
                for (int sampleI = 0; sampleI < samplesAtScatterI[scatterI]; ++sampleI) {
                    // Choose sample direction according to phase function
                    // TODO actually sample according to phaseHG
                    u1 = samp[sampOffsetStep + sampOffsetSample + 1];
                    u2 = samp[sampOffsetStep + sampOffsetSample + 2];
                    /*
                    wo = SampleHG (w, vr->getG(), u1, u2);
                    wo /= wo.Length(); // results were usually a bit off from normal (~.1)
                    // probability of o according to phase function
                    float phasePdf = HGPdf(w, -wo, vr->getG());
                    */
                    /*
                    wo = SampleCosWeightedSphere (ray.d, u1, u2);
                    float samplePdf = CosWeightedSpherePdf (ray.d, wo);
                    float phaseProb = max(vr->p (p, w, -wo), 1.f);
                    */
                    wo = UniformSampleSphere (u1, u2);
                    float samplePdf = INV_FOURPI;
                    float phaseProb = vr->p (p, w, -wo);

/*
                    cout << "ray (" << ray.d.x << "," << ray.d.y << "," << ray.d.z << ")\n"
                         << "wo  (" << wo.x << "," << w.y << "," << w.z << ")\n"
                         << "u1:             " << u1 << endl
                         << "dot (ray.d,wo): " << Dot(ray.d, wo) << endl
                         << "samplePdf     : " << samplePdf << endl
                         << "phaseProb     : " << phaseProb << "\n\n";
                         */
                    /*
                    if (phasePdf > 1)
                        cout << "ERROR PHASEPDF > 1\n";
                        */

                    RayDifferential r (p, wo);
                    Spectrum L = Li_multi (scene, r, NULL, alpha, scatterI + 1, 1.f);
                    
                    Ld += (L * phaseProb) / samplePdf;

                    sampOffsetSample += randPerSample;
                }
                Ld /= (float)samplesAtScatterI[scatterI];
                Lv += Tr * ss * Ld;
            }

            // we're in the final scatter, compute single scattering from light
            // TODO multiple samples for non-delta lights
            else {
                // randomly choose one of the lights to sample
		        int lightNum = min (Floor2Int (samp[sampOffsetStep + 0 + 0] * nLights),
			                        nLights - 1);
		        Light *light = scene->lights[lightNum];

		        // Add contribution of _light_ due to scattering at _p_
		        float pdf;
		        VisibilityTester vis;
                u1 = samp[sampOffsetStep + 0 + 1];
                u2 = samp[sampOffsetStep + 0 + 2];
		        Spectrum L = light->Sample_L(p, u1, u2, &wo, &pdf, &vis);
		        if (!L.Black() && pdf > 0.f && vis.Unoccluded(scene)) {
                    Spectrum Ld (0.f);
                    Ld = L * vis.Transmittance(scene); // amount of light that got to this point along scattered ray

			        Lv += Ld // amount of light that got to this point
                          * ss * vr->p(p, w, -wo) // amount which scatters towards our eye
                          * Tr // amount which ever reaches our eye
                          * float(nLights) / pdf; // pretend this much light is coming from all possible sources;
		        }
            }
        }
	    sampOffsetStep += samplesAtScatterI[scatterI] * randPerSample;
	}
	return Lv * step;
}
extern "C" DLLEXPORT VolumeIntegrator *CreateVolumeIntegrator(const ParamSet &params) {
	float stepSize  = params.FindOneFloat("stepsize", 1.f);
    const int *samplesPerScatter;
    int nScatters;
    samplesPerScatter = params.FindInt("samplesPerScatter", &nScatters);
    if (samplesPerScatter == NULL) {
        Error ("No \"samplesPerScatter\" values provided");
        return NULL;
    }
	return new MultipleScattering(stepSize, nScatters, samplesPerScatter);
}
