#include "pbrt.h"
#include "lights/atmosphereparticle.h"
#include <iostream>
#include <fstream>
#include <sstream>

using namespace std;

#define MAX_ATMOSPHERE_LAYERS 6

float RayleighParticle::Density(float height)
{
	return exp(-height*scaleHeight);
}
// Measurements using aereal perspective indicate 
// ScatteringCoeff for 560nm is about 0.011818 km^-1
SpectrumHR RayleighParticle::ScatteringCoeff(float height)
{
	if (scatterCoeff.Black())
	{
		const float* wl;
		float samples[COLOR_SAMPLES_HR];
		wl = SpectrumHR::SampleWavelengths();
		for (int i = 0; i < COLOR_SAMPLES_HR; ++i)
		{
			samples[i] = powf(scatterConstant/wl[i], 4.f);
		}
		scatterCoeff = SpectrumHR(samples);
	}
	return Density(height) * scatterCoeff;
}

float MieParticle::Density(float height)
{
	return expf(-height*scaleHeight);
}
SpectrumHR MieParticle::ScatteringCoeff(float height)
{
	if (scatterCoeff.Black())
	{
		float samples[COLOR_SAMPLES_HR];
		const float* wl = SpectrumHR::SampleWavelengths();
		for (int i = 0; i < COLOR_SAMPLES_HR; ++i) {
			samples[i] = scatterConstant * powf(wl[i], -0.84f);
		}
		scatterCoeff = SpectrumHR(samples);
	}
	return Density(height) * scatterCoeff;
}
// OPACParticle definitions
OPACParticle::OPACParticle(const char* file)
{
	ifstream in;
	in.open(file);
	if (!in) {
		Error("Could not open file %s", file); 
	}
	string s;
	int line = 0;
	int layr = 0;
	string filenames[MAX_ATMOSPHERE_LAYERS];
	while (line >= 0) {
		if (!getline(in, s)) {
			Error("Unexpected end of particle file");
			exit(0);
		}
		if (s.size() == 0 || s[0] == '#')
			continue;
		istringstream sin(s);
		switch (line) {
		case 0:
			sin >> nParticleLayers;
			if (nParticleLayers > MAX_ATMOSPHERE_LAYERS) {
				nParticleLayers = MAX_ATMOSPHERE_LAYERS;
				Warning("More than %i layers specified", MAX_ATMOSPHERE_LAYERS);
			}
			cout << "gotnlayers\n";
			particleLayers = new ParticleLayer[nParticleLayers];
			++line;
			break;
		case 1:
			sin >> filenames[layr];
			sin >> particleLayers[layr].top;
			sin >> particleLayers[layr].p0;
			sin >> particleLayers[layr].Zinv;
			//cout << "gotlayer: " << filenames[layr] << "\n";
			++layr;
			if (layr >= nParticleLayers)
				line = -1;
			break;
		}
	}
	for (int i = 0; i < nParticleLayers; ++i) {
		ReadLayer(filenames[i].c_str(), particleLayers + i);
	}
}
void OPACParticle::ReadLayer(const char* file, ParticleLayer* layer) {
	ifstream in;
	in.open(file);
	if (!in) {
		Error("Could not open file %s", file); 
	}
	string s;
	int wl = 0;
	float attenCoeff[COLOR_SAMPLES_HR];
	float scatterF[NUM_PHASE_SAMPLES][COLOR_SAMPLES_HR];
	const float* sampleWls = SpectrumHR::SampleWavelengths();
	while (wl < COLOR_SAMPLES_HR)
	{
		if (!getline(in, s))
			Error("Unexpected end of particle file");
		if (s.size() == 0 || s[0] == '#')
			continue;
		istringstream sin(s);
		float val;
		// Get the wavelength
		sin >> val;
		if (fabsf(val*1000.f - sampleWls[wl]) > 0.0001f)
			Error("File wavelength %f does not match sample wavelength %f", val*1000.f, sampleWls[wl]);
		// Get extinction coeff
		sin >> attenCoeff[wl];
		// Get stupid phase function line
		if (!getline(in, s))
			Error("Unexpected end of particle file");
		for (int i = 0; i < 14; ++i) {
			if (!getline(in, s))
				Error("Unexpected end of particle file");
			istringstream stream(s);
			for (int j = 0; j < 8; ++j) {
				stream >> scatterF[8*i + j][wl];
			}
		}
		++wl;
	}
	for (int i = 0; i < NUM_PHASE_SAMPLES; ++i) {
		layer->volumeScatteringFunction[i] = SpectrumHR(scatterF[i]);
	}
	layer->attenuationCoeff = SpectrumHR(attenCoeff);
}
static int binarySearch(float needle, const float* haystack, int bot, int top)
{
	if (bot == top)
		return bot;
	if (abs(bot-top) == 1)
		return bot;
	int split = (top-bot+1)/2+bot;
	if (needle < haystack[split])
		return binarySearch(needle, haystack, bot, split);
	else
		return binarySearch(needle, haystack, split, top);
}
ParticleLayer* OPACParticle::GetLayer(float height) {
	for (int i = 0; i < nParticleLayers; ++i) {
		if (height < particleLayers[i].top)
			return particleLayers + i;
	}
	return NULL;
}
SpectrumHR OPACParticle::VolumeScatterF(const Vector& w, const Vector& wp, float height) {
	float theta = Degrees(acosf(Dot(w, wp)));
	// Get the current layer
	ParticleLayer* layer = GetLayer(height);
	if (!layer)
		return SpectrumHR(0.f);

	if (theta > phaseSampleAngles[NUM_PHASE_SAMPLES - 1])
		return layer->volumeScatteringFunction[NUM_PHASE_SAMPLES - 1]*Density(height);
	
	// Find the two angles which it falls between
	int index = binarySearch(theta, phaseSampleAngles, 0, NUM_PHASE_SAMPLES-1);
	float weight = (theta - phaseSampleAngles[index]) /
				   (phaseSampleAngles[index+1]-phaseSampleAngles[index]);

	SpectrumHR diff = layer->volumeScatteringFunction[index+1] - layer->volumeScatteringFunction[index];
	SpectrumHR result = layer->volumeScatteringFunction[index];
	result.AddWeighted(weight, diff);

	return result*Density(height);
}
SpectrumHR OPACParticle::AttenuationCoeff(float height) {
	ParticleLayer* layer = GetLayer(height);
	if (!layer)
		return SpectrumHR(0.f);
	return Density(height, layer)*layer->attenuationCoeff;
}
float OPACParticle::Density(float height) {
	ParticleLayer* layer = GetLayer(height);
	if (!layer)
		return 0.f;
	if (layer->Zinv > 90.f || height < 0.f)
		return densityScale*layer->p0;
	return densityScale*layer->p0*exp(-height*layer->Zinv);
}
float OPACParticle::Density(float height, ParticleLayer* layer) {
	if (layer->Zinv > 90.f || height < 0.f)
		return densityScale*layer->p0;
	return densityScale*layer->p0*exp(-height*layer->Zinv);
}

const float OPACParticle::phaseSampleAngles[NUM_PHASE_SAMPLES] = {
	0.000f,    0.100f,    0.200f,    0.300f,    0.400f,    0.500f,    0.600f,    0.700f,
	0.800f,    0.900f,    1.000f,    1.100f,    1.200f,    1.300f,    1.400f,    1.500f,
	1.600f,    1.700f,    1.800f,    1.900f,    2.000f,    3.000f,    4.000f,    5.000f,
	6.000f,    8.000f,    10.000f,   12.000f,   14.000f,   16.000f,   18.000f,   20.000f,
	22.000f,   24.000f,   26.000f,   28.000f,   30.000f,   32.000f,   34.000f,   36.000f,
	38.000f,   40.000f,   42.000f,   44.000f,   46.000f,   48.000f,   50.000f,   52.000f,
	54.000f,   56.000f,   58.000f,   60.000f,   62.000f,   64.000f,   66.000f,   68.000f,
	70.000f,   72.000f,   74.000f,   76.000f,   78.000f,   80.000f,   82.000f,   84.000f,
	86.000f,   88.000f,   90.000f,   92.000f,   94.000f,   96.000f,   98.000f,   100.000f,
	102.000f,  104.000f,  106.000f,  108.000f,  110.000f,  112.000f,  114.000f,  116.000f,
	118.000f,  120.000f,  122.000f,  124.000f,  126.000f,  128.000f,  130.000f,  132.000f,
	134.000f,  136.000f,  138.000f,  140.000f,  142.000f,  144.000f,  146.000f,  148.000f,
	150.000f,  152.000f,  154.000f,  156.000f,  158.000f,  160.000f,  162.000f,  164.000f,
	166.000f,  168.000f,  170.000f,  172.000f,  174.000f,  176.000f,  178.000f,  180.000f
};
