#ifndef _OPAC_PARTICLE_H
#define _OPAC_PARTICLE_H

#include "pbrt.h"
#include "color.h"
#include "volume.h"

#define NUM_PHASE_SAMPLES 112

// Base class for different types of atmospheric particles
class AtmosphereParticle
{
public:
	virtual ~AtmosphereParticle() {};
	virtual float Density(float height)=0;
	virtual SpectrumHR AttenuationCoeff(float height)=0;
	virtual SpectrumHR VolumeScatterF(const Vector& w, const Vector& wp,
									  float height)=0;
};
class RayleighParticle : public AtmosphereParticle
{
public:
	RayleighParticle() {scatterCoeff = SpectrumHR(0.f);}
	float Density(float height);
	SpectrumHR AttenuationCoeff(float height) {return ScatteringCoeff(height);};
	SpectrumHR VolumeScatterF(const Vector& w, const Vector& wp, float height)
		{return ScatteringCoeff(height)*Phase(w, wp);};
	SpectrumHR ScatteringCoeff(float height);
	float Phase(const Vector& w, const Vector& wp)
		{return PhaseRayleigh(w, wp);};
private:
	SpectrumHR scatterCoeff;
	// parameter for particle density as function of height
	static const float scaleHeight = 0.055f;
	// scaling for scattering coefficient
	static const float scatterConstant = 800.f;
};
class MieParticle : public AtmosphereParticle
{
public:
	MieParticle() {scatterCoeff = SpectrumHR(0.f);}
	float Density(float height);
	SpectrumHR AttenuationCoeff(float height) {return ScatteringCoeff(height);};
	SpectrumHR ScatteringCoeff(float height);
	SpectrumHR VolumeScatterF(const Vector& w, const Vector& wp, float height)
		{return ScatteringCoeff(height)*Phase(w, wp);};
	float Phase(const Vector& w, const Vector& wp)
		{return PhaseHG(w, wp, 0.9f);};
private:
	SpectrumHR scatterCoeff;
	static const float scaleHeight = 0.125;
	static const float scatterConstant = 200.f;
};
typedef struct
{
	float top;
	float p0;
	float Zinv;
	SpectrumHR attenuationCoeff;
	SpectrumHR volumeScatteringFunction[NUM_PHASE_SAMPLES];
} ParticleLayer;
// Particles obtained from the mixture descriptions in the OPAC software package
class OPACParticle : public AtmosphereParticle {
public:
	OPACParticle(const char *file);
	SpectrumHR VolumeScatterF(const Vector& w, const Vector& wp, float height);
	SpectrumHR AttenuationCoeff(float height);t
	float Density(float height);
private:
	// private methods
	void ReadLayer(const char* file, ParticleLayer* layer);
	ParticleLayer* GetLayer(float height);
	float Density(float height, ParticleLayer* layer);
	// private data
	ParticleLayer* particleLayers;
	int nParticleLayers;
	const static float phaseSampleAngles[NUM_PHASE_SAMPLES];
	const static float densityScale = 0.25f;
};
#endif
