
/*
 *
 */

// sky.cpp*
#include "pbrt.h"
#include "light.h"
#include "texture.h"
#include "shape.h"
#include "scene.h"
#include "mipmap.h"

#include "volume.h"
#include "lights/atmosphereparticle.h"
#include <iostream>
#include <fstream>
#include <sstream>

using namespace std;

// Awesome constants
#define RADIUS_OF_EARTH_KM 6378.1f
#define DEPTH_OF_ATMOSPHERE_KM 50.f
#define TEMPERATURE_OF_SUN_K 6000.f
// Debug constants
//#define EARTH_SHADOW
#define HR_SPECTRUM

// Function
extern float PhaseMieHazy(const Vector& w, const Vector& wp);
extern float PhaseRayleigh(const Vector& w, const Vector& wp);
extern SpectrumHR Exp(const SpectrumHR& s);
extern ostream &operator<<(ostream &, const SpectrumHR &);
// Intersect Ray Sphere
static bool RaySphereIntersect(const Ray &r, Point *pHit, float *tHit, float radius)
 {
	// Compute quadratic sphere coefficients
	float A = r.d.x*r.d.x + r.d.y*r.d.y +
			r.d.z*r.d.z;
	float B = 2 * (r.d.x*r.o.x + r.d.y*r.o.y +
			r.d.z*r.o.z);
	float C = r.o.x*r.o.x + r.o.y*r.o.y +
			r.o.z*r.o.z - radius*radius;
	// Solve quadratic equation for _t_ values
	float t0, t1;
	if (!Quadratic(A, B, C, &t0, &t1))
	{
		return false;
	}
	// Compute correct time of hit with sphere
	float thit = t0;
	if (thit < 0.f)
	{
		if (t1 < 0.f)
			return false;
		else
			thit = t1;
	}
	if (tHit != NULL)
		*tHit = thit;
	// Compute sphere hit position
	if (pHit != NULL)
		*pHit = r(thit);
	return true;
}
// PBSkyLight Declarations
class PBSkyLight : public Light {
public:
	// PBSkyLight Public Methods
	PBSkyLight(const Transform &light2world, int ns, Vector sun,
			   Vector up, int intSteps, int odSteps);
	PBSkyLight(const Transform &light2world, int ns, float sunTheta,
			   float sunPhi, Vector up, Vector north, int intSteps, int odSteps);
	Spectrum Power(const Scene *scene) const {
		return Spectrum(0.f);
	}
	bool IsDeltaLight() const { return false; }
	Spectrum Le(const RayDifferential &r) const;
	Spectrum Sample_L(const Point &p, float u1, float u2, Vector *wi, float *pdf,
		VisibilityTester *visibility) const;
	Spectrum Sample_L(const Scene *scene, float u1, float u2,
			float u3, float u4, Ray *ray, float *pdf) const;
	float Pdf(const Point &, const Vector &) const;
	Spectrum Sample_L(const Point &P, Vector *w, VisibilityTester *visibility) const;
private:
	// PBSkyLight Private Methods
	SpectrumHR Transmission(Point p) const;
	SpectrumHR OpticalLengthS(Point p) const;
	SpectrumHR ScatterWeight(const Point& p, const Vector& wv, const Vector& ws) const;
	void Init(int intSteps, int odSteps);
	SpectrumHR SolarSpectrum();
	// PBSkyLight Private Data
	int nParticles;
	AtmosphereParticle** particles;
	SpectrumHR solarRadiance;
	float atmosphereDepth;
	int numSteps;
	int numODSteps;
	Vector sunDir;
	Vector zenith;
	Point surfacePos;
	float LScale;
	float tauScale;
	mutable bool debugBool;
};
// We model the sun as a black body with temerature around 7000 degrees Kelvin
SpectrumHR PBSkyLight::SolarSpectrum()
{
	float scalefactor = 12.f;
	const float* wl = SpectrumHR::SampleWavelengths();
	float samples[COLOR_SAMPLES_HR];
	for (int i = 0; i < COLOR_SAMPLES_HR; ++i)
	{
		float A = 1.f / (expf(0.0143877506f/(wl[i]*(0.000000001f)*TEMPERATURE_OF_SUN_K)) - 1);
		samples[i] = A * (scalefactor / powf(wl[i]*0.001,5));
	}
	return SpectrumHR(samples);
}
void PBSkyLight::Init(int intSteps, int odSteps){
	atmosphereDepth = DEPTH_OF_ATMOSPHERE_KM;
	numSteps = intSteps;
	numODSteps = odSteps;
	nParticles = 2;
	particles = new AtmosphereParticle*[nParticles];
	//particles[0] = new OPACParticle("atmosphere/continentalPollutedMix");
	particles[0] = new RayleighParticle();
	particles[1] = new MieParticle();
	Vector surfaceVect = (zenith * RADIUS_OF_EARTH_KM) + (zenith * 0.001f);
	surfacePos = Point(surfaceVect.x, surfaceVect.y, surfaceVect.z);
	solarRadiance = SolarSpectrum();
	LScale = 0.0004f;
	// Correction factor for optical depth
	tauScale = 0.0014f;
	debugBool = false;
}
PBSkyLight::PBSkyLight(const Transform &light2world, int ns, Vector sun,
	Vector up, int intSteps, int odSteps)
	: Light(light2world, ns) {
	zenith = Normalize(up);
	sunDir = Normalize(sun);
	Init(intSteps, odSteps);
}
PBSkyLight::PBSkyLight(const Transform &light2world, int ns, float sunTheta,
	float sunPhi, Vector up, Vector north, int intSteps, int odSteps)
	: Light(light2world, ns) {
	zenith = Normalize(up);
	Vector east = Cross(north, up);
	// Compute sun vector from solar angles
	Transform pitch = Rotate(sunPhi, east);
	Transform yaw = Rotate(sunTheta, north);
	sunDir = north;
	sunDir = pitch(sunDir);
	sunDir = yaw(sunDir);
	Init(intSteps, odSteps);
}
SpectrumHR PBSkyLight::Transmission(Point p) const{
	SpectrumHR T(0.f);
	float height = (p - Point(0.f, 0.f, 0.f)).Length() - RADIUS_OF_EARTH_KM;
	if (height < 0.f)
		height = 0.f;
	for (int i = 0; i < nParticles; ++i)
	{
		T += particles[i]->AttenuationCoeff(height);
	}
	return T;
}
SpectrumHR PBSkyLight::OpticalLengthS(Point p) const{
	Ray pToSun(p, sunDir);
	Point pHitAtm;
	RaySphereIntersect(pToSun, &pHitAtm, NULL, RADIUS_OF_EARTH_KM + atmosphereDepth);
	Ray pToAtm(p, pHitAtm - p);
	float stepLength = 1.f / numODSteps;
	float stepLengthKM = (pToAtm(stepLength) - p).Length();
	// sample halfway through each step
	float t = 0.5f * stepLength;
	SpectrumHR tau(0.f);
	for (int i = 0; i < numODSteps; ++i)
	{
		tau += Transmission(pToAtm(t));
		t += stepLength;
	}
	return tau * stepLengthKM * tauScale;
}
Spectrum
	PBSkyLight::Le(const RayDifferential &r) const {
	// Assume that view position is on surface of earth
	Ray viewRay;
	viewRay.o = surfacePos;
	Point p = viewRay.o, pPrev;
	viewRay.d = r.d;
	if (RaySphereIntersect(viewRay, NULL, NULL, RADIUS_OF_EARTH_KM)) {
		viewRay.d.z = -viewRay.d.z;//return Spectrum(0.f);
	}
	// Time at which ray hits atmosphere
	Point pHitAtm;
	float t1;
	if (!RaySphereIntersect(viewRay, &pHitAtm, &t1, RADIUS_OF_EARTH_KM + atmosphereDepth))
		return Spectrum(0.f);

	SpectrumHR L_total(0.f);
	// Optical length to the viewer
	SpectrumHR opticalLengthV(0.f);
	float stepLength = (viewRay(t1) - viewRay.o).Length() / numSteps;
	float t0 = RandomFloat() * stepLength;
	// Numerically integrate first order scattering along view ray
	SpectrumHR L_1(0.f);
	for (int i = 0; i < numSteps; ++i, t0 += stepLength)
	{
		pPrev = p;
		p = viewRay(t0);
		Ray stepRay(pPrev, p-pPrev);
		opticalLengthV += stepLength * Transmission(stepRay(RandomFloat())) * tauScale;
		// Get first order radiance contribution at point p
#ifdef EARTH_SHADOW
		Ray earthShadowRay(p, sunDir);
		if (RaySphereIntersect(earthShadowRay, NULL, RADIUS_OF_EARTH_KM ))
			L_1 = SpectrumHR(0.0001f);
		else
			L_1 = solarRadiance*ScatterWeight(p, -viewRay.d, -sunDir)*Exp(-opticalLengthV - OpticalLengthS(p));
#else
		L_1 = solarRadiance*ScatterWeight(p, -viewRay.d, -sunDir)*Exp(-opticalLengthV - OpticalLengthS(p));
#endif
		L_total += L_1;
	}
	L_total *= stepLength;

	float xyz[3];
	L_total.XYZ(xyz);
	return FromXYZ(xyz[0],xyz[1],xyz[2]) * LScale;
}
SpectrumHR PBSkyLight::ScatterWeight(const Point& p, const Vector& wv, const Vector& ws) const {
	SpectrumHR weight(0.f);
	float height = (p - Point(0.f, 0.f, 0.f)).Length() - RADIUS_OF_EARTH_KM;
	if (height < 0.f)
		height = 0.f;
	for (int i = 0; i < nParticles; ++i)
	{
		weight += particles[i]->VolumeScatterF(Normalize(wv), Normalize(ws), height);
	}
	return weight;
}
Spectrum PBSkyLight::Sample_L(const Point &p,
		float u1, float u2, Vector *wi, float *pdf,
		VisibilityTester *visibility) const {
	*wi = UniformSampleHemisphere(u1, u2);
	*pdf = UniformHemispherePdf(0.f, 0.f);
	visibility->SetRay(p, *wi);
	return Le(RayDifferential(p, *wi));
}
float PBSkyLight::Pdf(const Point &p, const Vector &w) const {
	return UniformHemispherePdf(0.f,0.f);
}
// For photon mapping, etc...
Spectrum PBSkyLight::Sample_L(const Scene *scene,
		float u1, float u2, float u3, float u4,
		Ray *ray, float *pdf) const {
	return Spectrum(0.f);
}
Spectrum PBSkyLight::Sample_L(const Point &p,
		Vector *wi, VisibilityTester *visibility) const {
	*wi = UniformSampleHemisphere(RandomFloat(), RandomFloat());
	visibility->SetRay(p, *wi);
	return Le(RayDifferential(p, *wi));
}
#define FLOAT_NOT_SET -19999.9f
extern "C" DLLEXPORT Light *CreateLight(const Transform &light2world,
		const ParamSet &paramSet) {
	int nSamples = paramSet.FindOneInt("nsamples", 1);
	Vector zenith = paramSet.FindOneVector("zenith", Vector(0.f,0.f,1.f));
	Vector sunDir = paramSet.FindOneVector("sundir", Vector(0.f,0.f,1.f));
	Vector north = paramSet.FindOneVector("north", Vector(0.f,1.f,0.f));
	float sunTheta = paramSet.FindOneFloat("suntheta", FLOAT_NOT_SET);
	float sunPhi = paramSet.FindOneFloat("sunphi", 0.f);
	int intSteps = paramSet.FindOneInt("viewraysamples", 20);
	int odSteps = paramSet.FindOneInt("sunraysamples", 10);
	PBSkyLight* light;
	if (sunTheta == FLOAT_NOT_SET)
		light = new PBSkyLight(light2world, nSamples, sunDir, zenith, intSteps, odSteps);
	else
		light = new PBSkyLight(light2world, nSamples, sunTheta, sunPhi, zenith, north, intSteps, odSteps);
	return light;
}
