
/*
 * pbrt source code Copyright(c) 1998-2005 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For educational use only; commercial use expressly forbidden.
 * NO WARRANTY, express or implied, for this software.
 * (See file License.txt for complete license)
 */

// sky.cpp*
#include "pbrt.h"
#include "light.h"
#include "texture.h"
#include "shape.h"
#include "scene.h"
#include "mipmap.h"
// Function
extern Spectrum FromXYZ(float x, float y, float z);
// Matrix multiplication functions
static void matrixmult (int ARows, int ACols, int BCols,
						const float* A, const float* B, float* C)
{
	int i, j, k;
	memset(C, 0, ARows*BCols*sizeof(float));
	for (i = 0; i < ARows; ++i) {
		for (j = 0; j < BCols; ++j) {
			for (k = 0; k < ACols; ++k) {
				C[i*BCols + j] += A[i*ACols + k] * B[k*BCols + j];
			}
		}
	}
}
static void matrixmult (int ARows, int ACols, int BCols,
						const double* A, const double* B, double* C)
{
	int i, j, k;
	memset(C, 0, ARows*BCols*sizeof(double));
	for (i = 0; i < ARows; ++i) {
		for (j = 0; j < BCols; ++j) {
			for (k = 0; k < ACols; ++k) {
				C[i*BCols + j] += A[i*ACols + k] * B[k*BCols + j];
			}
		}
	}
}
// TODO get rid of this matrix test code
static void testMatrixMult()
{
	float A[6] = {2.f, 3.f, 1.f, 6.f, 4.f, 2.f};
	float B[4] = {1.f, 0.f, 3.f, 2.f};
	float actual[6] = {11.f, 6.f, 19.f, 12.f, 10.f, 4.f};
	float result[6] = {0.f, 0.f, 0.f, 0.f, 0.f, 0.f};
	matrixmult(3, 2, 2, A, B, result);
	for (int i = 0; i < 6; ++i)
	{
		if (actual[i] != result[i])
			Error("WRONG: %f != %f", result[i], actual[i]);
	}
	Info("MatrixMult test complete");
}
// SkyLight Declarations
class SkyLight : public Light {
public:
	// SkyLight Public Methods
	SkyLight(const Transform &light2world, int ns, Vector up, Vector sun, float turb);
	Spectrum Power(const Scene *scene) const {
		return Spectrum(0.f);
	}
	bool IsDeltaLight() const { return false; }
	Spectrum Le(const RayDifferential &r) const;
	Spectrum Sample_L(const Point &p, const Normal &n,
		float u1, float u2, Vector *wi, float *pdf,
		VisibilityTester *visibility) const;
	Spectrum Sample_L(const Point &p, float u1, float u2, Vector *wi, float *pdf,
		VisibilityTester *visibility) const;
	Spectrum Sample_L(const Scene *scene, float u1, float u2,
			float u3, float u4, Ray *ray, float *pdf) const;
	float Pdf(const Point &, const Normal &, const Vector &) const;
	float Pdf(const Point &, const Vector &) const;
	Spectrum Sample_L(const Point &P, Vector *w, VisibilityTester *visibility) const;
private:
	// SkyLight Private Methods
	void ComputeDistributionCoefficients();
	// SkyLight Private Data
	float turbidity;
	Vector sunDir;
	Vector zenith;
	float zenithAngle;
	float YSkylightDistCoeffs[5];
	float xSkylightDistCoeffs[5];
	float ySkylightDistCoeffs[5];
	float Y_c;
	float x_c;
	float y_c;
};
static inline float PerezFunction(const float *coeffs, float theta, float gamma)
{
	return (1.f + coeffs[0]*expf(coeffs[1]/cosf(theta)))*
			(1.f + coeffs[2]*expf(coeffs[3]*gamma) + coeffs[4]*cosf(gamma)*cosf(gamma));
}
// Reference: http://www.imageval.com/public/Products/ISET/Manual/ProgrammersManual/ISET/color/index.html
static inline void YxyToXYZ(float Y, float x, float y, float* XYZ)
{
	XYZ[0] = (x/y)*Y;
	XYZ[1] = Y;
	XYZ[2] = ((1.f - x - y)/y) * Y;
}
SkyLight::SkyLight(const Transform &light2world, int ns, Vector up, Vector sun, float turb)
	: Light(light2world, ns) {
	turbidity = turb;
	sunDir = Normalize(sun);
	zenith = Normalize(up);
	zenithAngle = acosf(Dot(sunDir, zenith));
	// Compute zenith luminance
	float X = (0.44444444f - turbidity/120.f)*(M_PI - 2*zenithAngle);
	float Y_zenith = (4.0453f*turbidity - 4.9710f)*tanf(X) - 0.2155f*turbidity + 2.4192f;
	float tm[3];
	tm[0] = turbidity*turbidity;
	tm[1] = turbidity;
	tm[2] = 1.f;
	float sm[4];
	sm[3] = 1.f;
	sm[2] = zenithAngle;
	sm[1] = zenithAngle*zenithAngle;
	sm[0] = sm[1]*zenithAngle;
	float xM[12] =
	{  0.0017f, -0.0037f,  0.0021f,  0.0000f,
	  -0.0290f,  0.0638f, -0.0320f,  0.0039f,
	   0.1169f, -0.2120f,  0.0605f,  0.2589f  };
	float yM[12] =
	{  0.0028f, -0.0061f,  0.0032f,  0.0000f,
	  -0.0421f,  0.0897f, -0.0415f,  0.0052f,
	   0.1535f, -0.2676f,  0.0667f,  0.2669f  };
	float xMsm[3];
	matrixmult(3, 4, 1, xM, sm, xMsm);
	float x_zenith, y_zenith;
	matrixmult(1, 3, 1, tm, xMsm, &x_zenith);
	float yMsm[3];
	matrixmult(3, 4, 1, yM, sm, yMsm);
	matrixmult(1, 3, 1, tm, yMsm, &y_zenith);
	// Compute distribution coefficients
	ComputeDistributionCoefficients();
	// Precompute Perez formulation for zenith luminance
	Y_c = Y_zenith / PerezFunction(YSkylightDistCoeffs, 0.f, zenithAngle);
	x_c = x_zenith / PerezFunction(xSkylightDistCoeffs, 0.f, zenithAngle);
	y_c = y_zenith / PerezFunction(ySkylightDistCoeffs, 0.f, zenithAngle);
}
void SkyLight::ComputeDistributionCoefficients()
{
	float turbidityMatrix[2];
	turbidityMatrix[0] = turbidity;
	turbidityMatrix[1] = 1.f;
	float YCoeffMatrix[10] =
	{  0.1787f, -1.4630f,
	  -0.3554f,  0.4275f,
	  -0.0227f,  5.3251f,
	   0.1206f, -2.5771f,
	  -0.0670f,  0.3703f
	};
	matrixmult(5, 2, 1, YCoeffMatrix,
			   turbidityMatrix, YSkylightDistCoeffs);
	float xCoeffMatrix[10] =
	{  -0.0193f, -0.2592f,
	   -0.0665f,  0.0008f,
	   -0.0004f,  0.2125f,
	   -0.0641f, -0.8989f,
	   -0.0033f,  0.0452f
	};
	matrixmult(5, 2, 1, xCoeffMatrix,
			   turbidityMatrix, xSkylightDistCoeffs);
	float yCoeffMatrix[10] =
	{  -0.0167f, -0.2608f,
	   -0.0950f,  0.0092f,
	   -0.0079f,  0.2102f,
	   -0.0441f, -1.6537f,
	   -0.0109f,  0.0529f
	};
	matrixmult(5, 2, 1, yCoeffMatrix,
			   turbidityMatrix, ySkylightDistCoeffs);
}
Spectrum
	SkyLight::Le(const RayDifferential &r) const {
	Vector normDir = Normalize(r.d);
	float theta = acosf(Dot(normDir, zenith));
	if (theta > M_PI/2.f)
		return Spectrum(0.f);
	float gamma = acosf(Dot(normDir, sunDir));
	// Get luminance and chromacity of sky
	float Y = Y_c * PerezFunction(YSkylightDistCoeffs, theta, gamma);
	float x = x_c * PerezFunction(xSkylightDistCoeffs, theta, gamma);
	float y = y_c * PerezFunction(ySkylightDistCoeffs, theta, gamma);
	// Convert xyY to XYZ
	float XYZ[3];
	YxyToXYZ(Y, x, y, XYZ);
	return FromXYZ(XYZ[0], XYZ[1], XYZ[2]);
}
Spectrum SkyLight::Sample_L(const Point &p,
		const Normal &n, float u1, float u2,
		Vector *wi, float *pdf,
		VisibilityTester *visibility) const {
	return Spectrum(0.f);
}
float SkyLight::Pdf(const Point &, const Normal &n,
		const Vector &wi) const {
	return AbsDot(n, wi) * INV_TWOPI;
}
Spectrum SkyLight::Sample_L(const Point &p,
		float u1, float u2, Vector *wi, float *pdf,
		VisibilityTester *visibility) const {
	return Spectrum(0.f);
}
float SkyLight::Pdf(const Point &, const Vector &) const {
	return 1.f / (4.f * M_PI);
}
Spectrum SkyLight::Sample_L(const Scene *scene,
		float u1, float u2, float u3, float u4,
		Ray *ray, float *pdf) const {
	return Spectrum(0.f);
}
Spectrum SkyLight::Sample_L(const Point &p,
		Vector *wi, VisibilityTester *visibility) const {
	return Spectrum(0.f);
}
extern "C" DLLEXPORT Light *CreateLight(const Transform &light2world,
		const ParamSet &paramSet) {
	int nSamples = paramSet.FindOneInt("nsamples", 1);
	Vector zenith = paramSet.FindOneVector("zenith", Vector(0.f, 0.f, 1.f));
	Vector sunDir = paramSet.FindOneVector("sundir", zenith);
	float turbidity = paramSet.FindOneFloat("turbidity", 1.f);
	return new SkyLight(light2world, nSamples, zenith, sunDir, turbidity);
}
